/**
 *  -----------------------------------------------------------------------------
 *  
 *  Holo-Edit, spatial sound trajectories editor, part of Holophon
 *  Copyright (C) 2006 GMEM
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *  
 *  -----------------------------------------------------------------------------
 */
package holoedit.gui;

import holoedit.HoloPlayer;
import holoedit.data.HoloWaveForm;
import holoedit.fileio.WaveFormReader;
import holoedit.fileio.WaveFormSimpleReader;
import holoedit.util.Ut;
import java.io.File;
import java.util.Vector;
import javax.swing.JFileChooser;

public class SimpleSoundPool
{
	private HoloPlayer hpRef;
	public boolean fine = false;
	public boolean done = false;
	public int error = WaveFormReader.NO_ERROR;
	public String errorFileName;
	public HoloWaveForm last;
	private Vector<File> unfound;
	private File soundFolder;
	private boolean choosingfolder;
	
	public SimpleSoundPool(HoloPlayer owner)
	{
		hpRef = owner;
		unfound = new Vector<File>();
	}

	public void clear()
	{
		hpRef.gestionPistes.soundPool.clear();
		unfound.clear();
	}

	public void importSound(File f, boolean verboseError)
	{
		if(f.getName().indexOf(' ') == -1)
		{
			if(f.getName().length() <= 32)
			{
				Thread t = new Thread(new SoundFileImporter(f, verboseError));
				t.setPriority(Thread.MAX_PRIORITY);
				t.setName("SF-ImportDeamon");
				t.setDaemon(true);
				t.start();
			} else {
				if(verboseError)
					Ut.alert("Import Error", "Please rename the soundfile, for compatibility reasons with Max/MSP,\nit cannot be longer than 32 characters (including the extension).");
				error = WaveFormReader.PATH_ERROR;
				done = true;
				fine = false;
			}
		} else {
			if(verboseError)
				Ut.alert("Import Error", "Please rename the soundfile, for compatibility reasons with Max/MSP,\nit cannot contain blank spaces.");
			error = WaveFormReader.PATH_ERROR;
			done = true;
			fine = false;
		}
	}

	class SoundFileImporter implements Runnable
	{
		File f;
		boolean verbose = true;

		public SoundFileImporter(File _f, boolean verb)
		{
			f = _f;
			verbose = verb;
		}

		public void run()
		{
			if (f != null && ((f.exists() && f.isFile())))
			{
				done = false;
				fine = false;
				if (!hpRef.gestionPistes.soundPool.contains(new HoloWaveForm(f)))
				{
					WaveFormSimpleReader wfr = new WaveFormSimpleReader(f);
					try
					{
						while (!wfr.isDone())
						{
							Thread.sleep(250);
//							System.out.println("soundpool-import-waiting");
//							System.out.print('.');
						}
//						System.out.println();
					}
					catch (InterruptedException e)
					{
						e.printStackTrace();
					}
					done = true;
					if (wfr.isFine())
					{
						HoloWaveForm[] arr = wfr.getHoloWaveForms();
						for (HoloWaveForm hwf : arr)
							hpRef.gestionPistes.soundPool.add(hwf);
						last = arr[arr.length-1];
						fine = true;
					} else {
						error = wfr.getError();
						errorFileName = wfr.soundFile.getName();
						if(verbose)
							switch(error)
							{
							case WaveFormReader.NO_ERROR :
								break;
							case WaveFormReader.MONO_ERROR :
								HoloPlayer.error("Only mono file for instance,\nOnly the left channel has been read.");
								break;
							case WaveFormReader.TYPE_ERROR :
								HoloPlayer.error(errorFileName + " is not a supported file type.");
								break;
							case WaveFormReader.PATH_ERROR :
								HoloPlayer.error("Please rename the soundfile, for compatibility reasons with Max/MSP,\nit cannot be longer than 32 characters (including the extension).");
								break;
							case WaveFormReader.PATH2_ERROR :
								HoloPlayer.error("Please rename the soundfile, for compatibility reasons with Max/MSP,\nit cannot contain blank spaces.");
								break;
							case WaveFormReader.SSDII_NORSRC_ERROR :
								HoloPlayer.error("This Sound Designer II file has no resource and cannot be read.");
								break;
							case WaveFormReader.SSDII_RSRC_ERROR :
								HoloPlayer.error("There are errors in this Sound Designer II resource file.");
								break;
							default :
								HoloPlayer.error("Problem while loading soundfile, aborted.");
								break;
							}
					}
					wfr.stop();
					wfr = null;
					System.runFinalization();
				} else {
					done = true;
					fine = true;
					last = new HoloWaveForm(f);
				}
			} else {
				done = true;
				fine = false;
				error = WaveFormReader.FILE_ERROR;
				unfound.add(f);
				if(!choosingfolder)
				{
					choosingfolder = true;
					HoloPlayer.error(f.getName() + " can't be found,\neven it's cached waveform.\nLocalize ... ");
				} else {
					try{
						while(choosingfolder)
						{
//							System.out.println("soundpool-folder-localize-waiting");
							Thread.sleep(500);
						}
					} catch (InterruptedException e){}
				}
				if(soundFolder != null)
				{
					File[] sounds = soundFolder.listFiles();
					for (int k = unfound.size() - 1; k >= 0; k--)
					{
						File fil = unfound.get(k);
						for (int i = 0; i < sounds.length; i++)
						{
							if (sounds[i].getName().equalsIgnoreCase(fil.getName()))
							{
								hpRef.soundPool.importSound(sounds[i], true);
								unfound.remove(fil);
							}
						}
					}		
				} else {
					JFileChooser chooser = new JFileChooser(f.getParentFile().exists() ? f.getParent() : hpRef.gestionPistes.holoDirectory);
					chooser.setAcceptAllFileFilterUsed(false);
					chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					chooser.setDialogTitle("Choose a folder for missing sounds");
					chooser.showDialog(null, "Choose this folder for missing sounds");
					if (chooser.getSelectedFile() != null)
					{
						File tmp = chooser.getSelectedFile();
						if (tmp.exists() && tmp.isDirectory())
						{
							soundFolder = tmp;
							choosingfolder = false;
							File[] sounds = soundFolder.listFiles();
							for (int k = unfound.size() - 1; k >= 0; k--)
							{
								File fil = unfound.get(k);
								for (int i = 0; i < sounds.length; i++)
								{
									if (sounds[i].getName().equalsIgnoreCase(fil.getName()))
									{
										hpRef.soundPool.importSound(sounds[i], true);
										unfound.remove(fil);
									}
								}
							}		
						}
					}
				}
				if(!unfound.isEmpty())
					HoloPlayer.error(f.getName() + " can't still be found, aborted.");
			}
		}
	}

}
