/**
 *  -----------------------------------------------------------------------------
 *  
 *  Holo-Edit, spatial sound trajectories editor, part of Holophon
 *  Copyright (C) 2006 GMEM
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *  
 *  -----------------------------------------------------------------------------
 */
package holoedit.gui;

import holoedit.Main;
import holoedit.data.HoloWaveForm;
import holoedit.opengl.OpenGLUt;
import holoedit.opengl.TessCallback;
import java.awt.GridLayout;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;
import javax.media.opengl.glu.GLUtessellator;
import javax.swing.JPanel;

public class WaveFormRenderer extends JPanel implements GLEventListener
{
	public float[] bgColor = {1,1,1,1};
	public float[] marginColor = {0.8f,0.8f,0.8f,1};
	public float[] waveColor = {0,0,0,1};
	public float[] polyWaveColor = {0.5f,0.5f,0.5f,0.5f};
	private HoloWaveForm hwf;
	private GLU glu = new GLU();
	private GL gl;
	private GLUtessellator tobj;
	private TessCallback tessCallback;
	private int width;
	private int height;
	private float margin = 0.02f;
	private float max = 1;
	private int listID = 0;
	private GLCanvas glp;
	
	public WaveFormRenderer(Main owner)
	{
		super();
		glp = new GLCanvas(owner.glcap, null, owner.glpb.getContext(), null);
		glp.addGLEventListener(this);
		setLayout(new GridLayout(1, 1));
		add(glp);
	}
	
	public void setHoloWaveForm(HoloWaveForm _hwf)
	{
		hwf = _hwf;
		if(hwf != null) max = hwf.getFileLength();
		else max = 1;
		glp.display();
	}
	
	public HoloWaveForm getHoloWaveForm()
	{
		return hwf;
	}

	public void init(GLAutoDrawable drawable)
	{
		gl = drawable.getGL();
		
		tessCallback = new TessCallback(gl,glu);
		tobj = glu.gluNewTess();
		glu.gluTessCallback(tobj, GLU.GLU_TESS_VERTEX, tessCallback);
		glu.gluTessCallback(tobj, GLU.GLU_TESS_BEGIN, tessCallback);
		glu.gluTessCallback(tobj, GLU.GLU_TESS_END, tessCallback);
		glu.gluTessCallback(tobj, GLU.GLU_TESS_ERROR, tessCallback);
		glu.gluTessCallback(tobj, GLU.GLU_TESS_COMBINE, tessCallback);

		gl.glShadeModel(GL.GL_SMOOTH); // Enable Smooth Shading
		gl.glClearColor(bgColor[0],bgColor[1],bgColor[2],bgColor[3]); // White Background
		gl.glPointSize(5);
		gl.glViewport(0, 0, width, height);
		gl.glEnable(GL.GL_POINT_SMOOTH);
		gl.glEnable(GL.GL_BLEND);
		gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glMatrixMode(GL.GL_PROJECTION);

		if(listID == 0)
		{
//			System.out.println("init list");
			listID = gl.glGenLists(1000);
		}
	}
	
	public void display(GLAutoDrawable drawable)
	{
		gl = drawable.getGL();
		gl.glClear(GL.GL_COLOR_BUFFER_BIT);
		gl.glLoadIdentity();
		
		glu.gluOrtho2D(0, max, 0, 1 + 2 * margin);
		drawMargins();
		
		if(hwf != null)
			if(hwf.isFine())
			{
				gl.glPushMatrix();
				listID = hwf.drawSoundPool(gl,margin,waveColor,listID);
				gl.glPopMatrix();
			}
		
		gl.glFlush();
	}

	private void drawMargins()
	{
		OpenGLUt.glColor(gl,marginColor);
		gl.glBegin(GL.GL_POLYGON);
			OpenGLUt.drawPoint(gl,0,1+margin);
			OpenGLUt.drawPoint(gl,max,1+margin);
			OpenGLUt.drawPoint(gl,max,1+2*margin);
			OpenGLUt.drawPoint(gl,0,1+2*margin);
		gl.glEnd();
		
		gl.glBegin(GL.GL_LINES);
			OpenGLUt.drawPoint(gl,0,0.5f+margin);
			OpenGLUt.drawPoint(gl,max,0.5f+margin);
		gl.glEnd();
		
		gl.glBegin(GL.GL_POLYGON);
			OpenGLUt.drawPoint(gl,0,0);
			OpenGLUt.drawPoint(gl,max,0);
			OpenGLUt.drawPoint(gl,max,margin);
			OpenGLUt.drawPoint(gl,0,margin);
		gl.glEnd();
	}
	
	public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h)
	{
		width = w;
		height = h;
		glu.gluOrtho2D(0, max, 0, 1 + 2 * margin);
	}
	
	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged){}	
}
