/**
 *  -----------------------------------------------------------------------------
 *  
 *  Holo-Edit, spatial sound trajectories editor, part of Holophon
 *  Copyright (C) 2006 GMEM
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *  
 *  -----------------------------------------------------------------------------
 */
package holoedit.rt;

import holoedit.Main;
import holoedit.data.HoloPoint;
import holoedit.data.HoloSpeaker;
import holoedit.data.HoloTrack;
import holoedit.data.HoloTraj;
import holoedit.data.HoloWaveForm;
import holoedit.data.WaveFormInstance;
import holoedit.util.Ut;
import java.awt.Color;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.Vector;
import com.illposed.osc.OSCListener;
import com.illposed.osc.OSCMessage;
import com.illposed.osc.OSCPortIn;
import com.illposed.osc.OSCPortOut;

public abstract class Connection implements Runnable, OSCListener
{
	public String PROTOCOL_VERSION = "0.1";
	protected static final boolean VERBOSE_IN = false;
	protected static final boolean VERBOSE_OUT = false;
	protected static final int ORDER_NONE = -1;
	protected static final int ORDER_STOP = 0;
	protected static final int ORDER_PLAY = 1;
	protected static final int ORDER_PAUSE = 2;
	protected static final int ORDER_RESUME = 3;
	protected static final int ORDER_RECORD = 4;
	protected boolean open = false;
	protected boolean autostop = false;
	protected boolean playing = false;
	protected boolean recording = false;
	protected boolean paused = false;
	protected boolean loop = false;
	protected int out = 13008;
	protected int in = 13005;
	protected String address = "localhost";
	protected OSCPortOut sender;
	protected OSCPortIn receiver;
	protected String keyOut = "/holospat";
	protected String keyIn = "/holoedit";
	protected Thread runner;
	protected int counter;
	protected int end;
	protected int beg;
	protected int total;
	protected boolean firstCue;
	protected Main mainRef;
	protected boolean looping = false;
	protected boolean preloading;
	protected long oldDate = -1, date = -1;
	protected int oversleep = 0;
	protected String currentEditName = "holo.edit";
	protected String currentSpatDomain = "";
	protected String currentSpatName = "";
	protected Vector<HoloTraj> recTrajs = new Vector<HoloTraj>();
	protected int recBegDate = 0;
	protected int loopNum = 0;
	protected int order; 
	protected HoloTrack currentTrack;

	public Connection(Main m)
	{
		mainRef = m;
	}
		
	public abstract void open();

	public abstract void close();

	protected abstract void newSender();
	
	protected abstract void newReceiver();	
	
	protected void treatPacket(Date msgDate, String key, String[] msg)
	{
		if(VERBOSE_IN)
		{
			System.out.print("OSCIN : "+key);
			for(String a:msg)
				System.out.print(" "+a);
			System.out.println();
		}
		if(key.equalsIgnoreCase(keyIn))
		{
			if (msg[0].equalsIgnoreCase("play"))
			{	
				order = ORDER_PLAY;
				play();
			}
			else if (msg[0].equalsIgnoreCase("resume"))
			{
				order = ORDER_RESUME;
				paused = true;
				pause();
			}
			else if (msg[0].equalsIgnoreCase("pause"))
			{
				order = ORDER_PAUSE;
				paused = false;
				pause();
			}
			else if (msg[0].equalsIgnoreCase("stop"))
			{
				order = ORDER_STOP;
				stop();
			}
			else if (msg[0].equalsIgnoreCase("record"))
			{
				order = ORDER_RECORD;
				record(Integer.parseInt(msg[1]) > 0);
			}
			else if (msg[0].equalsIgnoreCase("preload"))
				preload();
			else if (msg[0].equalsIgnoreCase("position"))
				position();
			else if (msg[0].equalsIgnoreCase("update"))
				spatUpdate();
			else if (msg[0].equalsIgnoreCase("time"))
				setTime(msg[1]);
			else if (msg[0].equalsIgnoreCase("begtime"))
				setBegTime(msg[1]);
			else if (msg[0].equalsIgnoreCase("endtime"))
				setEndTime(msg[1]);
			else if (msg[0].equalsIgnoreCase("totaltime"))
				setTotalTime(msg[1]);
			else if (msg[0].equalsIgnoreCase("loop"))
				setLoop(msg[1]);
			else if (msg[0].equalsIgnoreCase("holo.spat"))
			{	
				if (msg[1].equalsIgnoreCase("key"))
					keyOut = msg[2];
			} else if (msg[0].equalsIgnoreCase("querykey"))
				send(keyOut, "holo.edit key " + keyIn);
			else if (msg[0].equalsIgnoreCase("queryspeakers"))
				sendSpeakers();
			else if (msg[0].equalsIgnoreCase("tk"))
			{
				if(!msg[1].equalsIgnoreCase("#1"))
				{
					int i = Integer.parseInt(msg[1]);
					boolean o = Integer.parseInt(msg[2]) > 0;
					boolean found = false;
					for(int p = 0 ; p < mainRef.gestionPistes.getNbTracks() && !found ; p++)
						if(mainRef.gestionPistes.getTrack(p).getNumber()==i)
						{
							found = true;
							mainRef.gestionPistes.changeVisible(p);
							mainRef.gestionPistes.ts.checkVisible[p].check(o);
						}
					mainRef.room.display();
				}
			}
			else if (msg[0].equalsIgnoreCase("rec"))
				treatRecord(msg);
			else if (msg[0].equalsIgnoreCase("recenable"))
			{
				int tkNum = Integer.parseInt(msg[1]);
				boolean tkRecOn = Integer.parseInt(msg[2]) > 0;
				boolean found = false;
				for(int p = 0 ; p < mainRef.gestionPistes.getNbTracks() && !found ; p++)
				{
					HoloTrack tk = mainRef.gestionPistes.getTrack(p);
					if(tk.getNumber()==tkNum)
					{
						found = true;
						tk.setRecEnable(tkRecOn);
					}
				}
			}
			else 
				System.out.println("Holo-Edit doesn't understand "+msg[0]);
		}
	}
	
	protected void treatRecord(String[] msg)
	{
		int loopDur = end - beg;
		int tkNum = Integer.parseInt(msg[1]);
		float x = Float.parseFloat(msg[2]);
		float y = Float.parseFloat(msg[3]);
		float z = Float.parseFloat(msg[4]);
		boolean found = false;
		for(int p = 0 ; p < mainRef.gestionPistes.getNbTracks() && !found ; p++)
		{
			HoloTrack tk = mainRef.gestionPistes.getTrack(p);
			if(tk.getNumber()==tkNum && tk.isRecEnable())
			{
				found = true;
				recTrajs.get(p).add(new HoloPoint(x,y,z,Integer.parseInt(msg[5])+loopNum*loopDur,false));
			}
		}
	}
	
	protected void send(String key, String msg)
	{
		try
		{
			if(sender != null)
				sender.send(new OSCMessage(key,new Object[]{msg}));
			if(VERBOSE_OUT)
				System.out.println("OSCOUT : "+key+" "+msg);
		} catch (IOException e) {}
	}
	
	public void send(String msg)
	{
		send(keyOut,msg);
	}
	
	public void acceptMessage(Date msgDate, OSCMessage msg)
	{
		Object[] os = msg.getArguments();
		String[] s = new String[os.length];
		for(int i = 0 ; i < os.length ; i++)
			s[i] = os[i].toString();
		treatPacket(msgDate,msg.getAddress(),s);
	}

	public int getIn()
	{
		return in;
	}

	public void setIn(int in)
	{
		if(this.in != in)
		{
			this.in = in;
			newReceiver();
		}
	}

	public int getOut()
	{
		return out;
	}

	public void setOut(int out)
	{
		if(this.out != out)
		{
			this.out = out;
			newSender();
		}
	}
	
	public void setAddress(String a)
	{
		if(!this.address.equalsIgnoreCase(a))
		{
			try
			{
				InetAddress.getByName(address);
				address = a;
			} catch (UnknownHostException e)
			{
				address = "localhost";
				e.printStackTrace();
			}
			newSender();
		}
	}

	public void setKeyIn(String k)
	{
		if(!this.keyIn.equalsIgnoreCase(k))
		{
			keyIn = k;
			newReceiver();
		}
	}
	
	public String getKeyIn()
	{
		return keyIn;
	}
	
	public void setKeyOut(String k)
	{
		if(!this.keyOut.equalsIgnoreCase(k))
		{
			keyOut = k;
			newSender();
		}
	}
	
	public String getKeyOut()
	{
		return keyOut;
	}
	
	public void setTime(String msg)
	{
		try {
			int time = Integer.parseInt(msg);
			if(time >= 0)
			{
				boolean p = playing;
				if(p)
					stop();
				mainRef.score.cursorTime = (int)((float)time / 10);
				mainRef.score.display();
				if(p)
					play();
			}
		} catch (NumberFormatException e) {}
	}
	
	public void setTotalTime(String msg)
	{
		try {
			int time = Integer.parseInt(msg);
			if(time >= 0)
			{
				total = Ut.max(time,mainRef.counterPanel.getDate(3));
				mainRef.counterPanel.setCompteur2(3,total);
				mainRef.score.display();
			}
		} catch (NumberFormatException e) {}
	}
	
	public int getTotalTime()
	{
		return total;
	}
	
	public void setBegin(int b)
	{
		beg = b;
	}
	
	public void setBegTime(String msg)
	{
		try {
			int time = Integer.parseInt(msg);
			if(time >= 0)
			{
				boolean p = playing;
				if(p)
					stop();
				mainRef.counterPanel.setCompteur2(1,time);
				mainRef.room.display();
				if(p)
					play();
			}
		} catch (NumberFormatException e) {
		}
	}
	
	public void setEnd(int e)
	{
		end = e;
	}
	
	public void setEndTime(String msg)
	{
		try {
			int time = Integer.parseInt(msg);
			if(time >= 0)
			{
				boolean p = playing;
				if(p)
					stop();
				mainRef.counterPanel.setCompteur2(2,time);
				mainRef.room.display();
				if(p)
					play();
			}
		} catch (NumberFormatException e) {}
	}
	
	public int getCurrentTime()
	{
		return counter;
	}
	
	public void setCurrentTime(int c)
	{
		counter = c;
	}
	
	public void sendBegin()
	{
		if(open)
			send("begin "+mainRef.counterPanel.getDate(1));
	}
	
	public void sendEnd()
	{
		if(open)
			send("end "+mainRef.counterPanel.getDate(2));
	}
	
	public void sendTotal()
	{
		if(open)
			send("total "+mainRef.counterPanel.getDate(3));
	}
	
	public void setLoop(String msg)
	{
		try {
			int l = Integer.parseInt(msg);
			loop(l > 0);
		} catch (NumberFormatException e) {}
	}
	
	public boolean isOpen()
	{
		return open;
	}
	
	public boolean isPlaying()
	{
		return playing;
	}
	
	public boolean isRecording()
	{
		return recording;
	}
	
	public boolean getAutostop()
	{
		return autostop;
	}
	
	public void setAutostop(boolean a)
	{
		autostop = a;
	}
	
	public void record(boolean b)
	{
		if(open)
		{
			if(!b)
			{
				// stop recording
				for(int i = 0 ; i < recTrajs.size() ; i++)
				{
					HoloTraj ht = recTrajs.get(i);
					ht.remove(0);
				}
				for(HoloTrack tk:mainRef.gestionPistes.tracks)
					tk.update();
				recording = false;
				mainRef.transport.rec.setIcon(mainRef.transport.recOffState);
				if(order != ORDER_RECORD)
					send("record 0");
				else order = ORDER_NONE;
			} else {
				// start recording
				mainRef.gestionPistes.StoreToUndoAll();
				mainRef.modify();
				recTrajs = new Vector<HoloTraj>();
				if(playing)
				{	
					recBegDate = counter;
				} else {
					recBegDate = mainRef.score.cursorTime;
				}	
				for(int i = 0 ; i < mainRef.gestionPistes.getNbTracks() ; i++)
				{
					HoloTraj ht = new HoloTraj();
					currentTrack = mainRef.gestionPistes.getTrack(i);
					ht.add(new HoloPoint(0,0,0,recBegDate,true));
					recTrajs.add(ht);
					if(currentTrack.isVisible() && !currentTrack.isLocked() && currentTrack.isRecEnable())
					{
						currentTrack.addTraj(ht);
						currentTrack.update();
					}
				}
				recording = true;
				mainRef.transport.rec.setIcon(mainRef.transport.recOnState);
				if(order != ORDER_RECORD)
					send("record 1");
				else order = ORDER_NONE;
			}
		}
	}
	
	public void play()
	{
		if(open && !playing)
		{
			runner = new Thread(this);
			runner.setName("RT-player");
			runner.setPriority(Thread.MAX_PRIORITY);
			// PRELOAD
			preload();
			try
			{
				Thread.sleep(500);
			}
			catch (InterruptedException e)
			{
			}
			position();
			for (HoloTrack tk : mainRef.gestionPistes.tracks)
			{
				if (tk.waves.isEmpty())
					total = Ut.max(total, tk.getLastDate());
				else
					total = Ut.max(total, Ut.max(tk.getLastDate(), tk.waves.lastElement().getLastDate()));
			}
			if(order != ORDER_PLAY)
				send("play");
			else 
				order = ORDER_NONE;
			autostop = false;
			playing = true;
			firstCue = true;
			loopNum = 0;
			counter = mainRef.score.cursorTime;
			looping = !firstCue;
			oversleep = 0;
			runner.start();
			if (!mainRef.rtDisplay.isAnimating())
				mainRef.rtDisplay.start();
			mainRef.transport.start.setIcon(mainRef.transport.playState);
		}
	}
	
	public void stop()
	{
		for(HoloTrack tk:mainRef.gestionPistes.tracks)
			tk.stop(true);
		if(open && (playing || paused || recording))
		{
			if(runner != null)
				runner.interrupt();
			if(recording)
				record(false);
			if(order != ORDER_STOP)
				send("stop");
			else order = ORDER_NONE;
			recording = false;
			playing = false;
			paused = false;
			if(mainRef.rtDisplay.isAnimating())
				mainRef.rtDisplay.stop();
			mainRef.transport.start.setIcon(mainRef.transport.stopState);
			mainRef.transport.resume.setIcon(mainRef.transport.resumeState);
		}
		mainRef.room.display();
	}
	
	public void stopAndPlay()
	{
		// STOP
		runner.interrupt();
		send("jump");
		runner = new Thread(this);
		runner.setName("RT-player");
		runner.setPriority(Thread.MAX_PRIORITY);
		// PRELOAD
		preload();
		try
		{
			Thread.sleep(500);
		}
		catch (InterruptedException e) {}
		mainRef.transport.start.setIcon(mainRef.transport.playState);
		// PLAY
		for (HoloTrack tk : mainRef.gestionPistes.tracks)
		{
			if (tk.waves.isEmpty())
				total = Ut.max(total, tk.getLastDate());
			else
				total = Ut.max(total, Ut.max(tk.getLastDate(), tk.waves.lastElement().getLastDate()));
		}
		firstCue = true;
		counter = mainRef.score.cursorTime;
		looping = !firstCue;
		loopNum = 0;
		oversleep = 0;
		runner.start();
	}
	
	public void pause()
	{
		if(open)
		{
			if(paused)
			{
				if(order != ORDER_RESUME)
					send("resume");
				else order = ORDER_NONE;
				mainRef.transport.resume.setIcon(mainRef.transport.resumeState);
				paused = false;
			} else {
				if(order != ORDER_PAUSE)
					send("pause");
				else order = ORDER_NONE;
				mainRef.transport.resume.setIcon(mainRef.transport.pauseState);
				paused = true;
			}
		}
	}
	
	public void preload()
	{
		preloading = true;
		beg = mainRef.counterPanel.getDate(1);
		end = mainRef.counterPanel.getDate(2);
		if(open)
		{
			sendPaths();
			sendVisible();
			for(HoloTrack tk:mainRef.gestionPistes.tracks)
				for(int i = 1 ; i <= tk.waves.size() ; i++)
					tk.waves.get(i-1).preload(tk,i+100,mainRef.score.cursorTime, beg,end);
		}
		preloading = false;
	}
	
	public void spatUpdate()
	{
		String tmp = "tks "+mainRef.gestionPistes.getNbTracks();
		for(HoloTrack tk:mainRef.gestionPistes.tracks)
			tmp += " " + tk.getNumber();
		send(tmp);
		for(HoloTrack tk:mainRef.gestionPistes.tracks)
		{
			Color c = tk.getColor();
			String na = tk.getName();
			if(na.indexOf(' ') != -1)
				na = "\""+na+"\"";
			send("tkup "+tk.getNumber()+" "+na+" "+c.getRed()+" "+c.getGreen()+" "+c.getBlue());
		}
		position();
		preload();
	}
	
	public void loop(boolean l)
	{
		loop = l;
		if(open)
		{
			send("loop "+(loop ? 1 : 0));
			if (loop)
				mainRef.transport.loop.setIcon(mainRef.transport.loopOnState);
			else
				mainRef.transport.loop.setIcon(mainRef.transport.loopOffState);

		}
	}
	
	public boolean getLoop()
	{
		return loop;
	}
	
	public boolean changeLoop()
	{
		loop(!loop);
		return getLoop();
	}

	public void playstop()
	{
		if(open)
		{
			if(playing || autostop) stop();
			else play();
		}
	}

	public void position()
	{
		counter = mainRef.score.cursorTime;
		send("time "+counter);
		for(int i = 0, last = mainRef.gestionPistes.getNbTracks() ; i < last ; i++)
		{
			currentTrack = mainRef.gestionPistes.getTrack(i);
			currentTrack.stop(true);
			if(currentTrack.isVisible())
			{
				String vc = currentTrack.getLastValueAt(counter);
				if(!vc.equalsIgnoreCase("none"))
					send("tk "+currentTrack.getNumber()+" pos "+vc);
			}
		}
		mainRef.room.display();
	}
	
	public void sendVisible()
	{
		String tkVisi = "visible";
		for(HoloTrack t:mainRef.gestionPistes.tracks)
			tkVisi = tkVisi + " " + t.getNumber() + " " + (t.isVisible() ? 1 : 0);
		send(tkVisi);	
	}

	public void sendPaths()
	{
		Vector<String> paths = new Vector<String>();
		for(HoloWaveForm hwf:mainRef.gestionPistes.soundPool.getSounds())
		{	
			String p = hwf.getPath();
			if(!paths.contains(p))
				paths.add(p);
		}
		send("paths "+paths.size());
		for(int i = 0 ; i < paths.size() ; i++)
			send("path "+i+" "+paths.get(i));
	}
	
	public String getAddress()
	{
		return address;
	}
	
	public void run()
	{
		// si modif, modifier galement HoloPlayer
		try
		{
			while(playing)
			{
				while(preloading)
				{
					System.out.println("preloading...");
					Thread.sleep(10);
				}
				if(!paused)
				{
					oldDate = new Date().getTime();
					if(loop && counter == end+1)
					{
						send("looping");
						counter = beg;
						looping = true;
						loopNum++;
					} else if(counter > total && !recording){
						send("stop");
						playing = false;
						paused = false;
						autostop = true;
					}
						
					send("time "+counter);
					String c = "cues";
					for(int i = 0, last = mainRef.gestionPistes.getNbTracks() ; i < last ; i++)
					{
						currentTrack = mainRef.gestionPistes.getTrack(i);
						if(currentTrack.isVisible())
						{
							if((recording && !currentTrack.isRecEnable()) || !recording)
							{
								String vc = currentTrack.getValueAt(counter,autostop,looping);
								if(!vc.equalsIgnoreCase("none"))
									send("tk "+currentTrack.getNumber()+" pos "+vc);
							}
							for(WaveFormInstance w : currentTrack.waves)
							{	
								String wc = w.play(currentTrack, counter, firstCue, loop, looping);
								if(!wc.equalsIgnoreCase("none"))
									c = c + " " + wc;
							}
						}
					}
					if(!c.equalsIgnoreCase("cues"))
						send(c);
					firstCue = false;
					looping = false;
					counter ++;
					date = new Date().getTime();
					int delta = (int)(date - oldDate);
					if(delta < 10)
					{
						delta = 10 - delta;
						if(oversleep > delta)
						{
							oversleep -= delta;
						} else {
							delta -= oversleep;
							oversleep = 0;
							date = new Date().getTime();
							Thread.sleep(delta);
							oversleep += (int)(new Date().getTime() - date - delta);
						}
					}
				}
			}
		} catch (InterruptedException e) {}
		mainRef.transport.start.setIcon(mainRef.transport.stopState);
	}
	
	public String toString()
	{
		return "\t<osc portIn=\"" + mainRef.connection.getIn() + "\"" + " portOut=\"" + mainRef.connection.getOut() + "\""
		+ " address=\"" + mainRef.connection.getAddress() + "\"" + " open=\"" + mainRef.connection.isOpen() + "\""
		+ " keyIn=\"" + keyIn + "\"" + " keyOut=\"" + keyOut + "\"" + "/>\n";		
	}

	public void sendSpeakers()
	{
		HoloSpeaker sp;
		for(int i = 0 ; i < mainRef.gestionPistes.speakers.size() ; i++)
		{
			sp = mainRef.gestionPistes.speakers.get(i);
			send("spk "+(i+1)+" "+sp.X+" "+sp.Y+" "+sp.Z);
		}
	}
}